mod content;
mod doodad;
mod inputstatus;
mod physics;
mod render;
mod ship;
mod system;

use anyhow::Result;
use cgmath::{Deg, Point2};
use std::time::Instant;
use walkdir::WalkDir;
use winit::{
	event::{
		ElementState, Event, KeyboardInput, MouseButton, MouseScrollDelta, TouchPhase,
		VirtualKeyCode, WindowEvent,
	},
	event_loop::{ControlFlow, EventLoop},
	window::WindowBuilder,
};

use crate::{
	doodad::Doodad,
	inputstatus::InputStatus,
	physics::Pfloat,
	render::{GPUState, SpriteTexture},
	ship::{Ship, ShipKind},
	system::System,
};

pub const ZOOM_MIN: Pfloat = 200.0;
pub const ZOOM_MAX: Pfloat = 2000.0;

// Z-axis range for starfield stars
pub const STARFIELD_PARALLAX_MIN: f32 = 100.0;
pub const STARFIELD_PARALLAX_MAX: f32 = 200.0;
// Size of a square starfield tile, in game units.
// A tile of size PARALLAX_MAX * screen-size-in-game-units
// will completely cover a (square) screen. This depends on zoom!
//
// Use a value smaller than zoom_max for debug.
pub const STARFIELD_SIZE: u64 = STARFIELD_PARALLAX_MAX as u64 * ZOOM_MAX as u64;
// Average number of stars per game unit
pub const STARFIELD_DENSITY: f64 = 0.01;
// Number of stars in one starfield tile
// Must fit inside an i32
pub const STARFIELD_COUNT: u64 = (STARFIELD_SIZE as f64 * STARFIELD_DENSITY) as u64;

#[derive(Debug, Clone, Copy)]
struct Camera {
	/// Camera center
	pos: Point2<Pfloat>,

	/// Camera zoom
	/// (How many game units tall is the viewport?)
	zoom: Pfloat,
}

trait Spriteable {
	fn get_sprite(&self) -> Sprite;
}

struct Sprite {
	/// Name of the sprite to draw
	texture: SpriteTexture,

	/// This object's position, in world coordinates.
	pos: Point2<Pfloat>,

	/// The size of this sprite,
	/// given as height in world units.
	height: Pfloat,

	/// Scale factor.
	/// if this is 1, sprite height is exactly self.height.
	scale: Pfloat,

	/// This sprite's rotation
	/// (relative to north, measured ccw)
	angle: Deg<Pfloat>,

	/// Parallax factor.
	/// More positive => farther away
	/// More negative => closer
	/// Zero: no parallax.
	parallax: Pfloat,
}

struct Game {
	input: InputStatus,
	last_update: Instant,
	player: Ship,
	system: System,
	camera: Camera,
}

impl Game {
	fn new() -> Self {
		Game {
			last_update: Instant::now(),
			input: InputStatus::new(),
			player: Ship::new(ShipKind::Gypsum, (0.0, 0.0).into()),
			camera: Camera {
				pos: (0.0, 0.0).into(),
				zoom: 500.0,
			},
			system: System::new(),
		}
	}

	fn process_key(&mut self, state: &ElementState, key: &VirtualKeyCode) {
		self.input.process_key(state, key)
	}

	fn process_click(&mut self, state: &ElementState, key: &MouseButton) {
		self.input.process_click(state, key)
	}

	fn process_scroll(&mut self, delta: &MouseScrollDelta, phase: &TouchPhase) {
		self.input.process_scroll(delta, phase)
	}

	fn update(&mut self) {
		let t: Pfloat = self.last_update.elapsed().as_secs_f32();

		if self.input.key_thrust {
			self.player.body.thrust(50.0 * t);
		}

		if self.input.key_right {
			self.player.body.rot(Deg { 0: 35.0 } * t);
		}

		if self.input.key_left {
			self.player.body.rot(Deg { 0: -35.0 } * t);
		}

		if self.input.v_scroll != 0.0 {
			self.camera.zoom = (self.camera.zoom + self.input.v_scroll).clamp(ZOOM_MIN, ZOOM_MAX);
			self.input.v_scroll = 0.0;
		}

		self.player.body.tick(t);
		self.camera.pos = self.player.body.pos;

		self.last_update = Instant::now();
	}

	fn get_sprites(&self) -> Vec<Sprite> {
		let mut sprites: Vec<Sprite> = Vec::new();

		sprites.append(&mut self.system.get_sprites());
		sprites.push(self.player.get_sprite());

		// Make sure sprites are drawn in the correct order
		// (note the reversed a, b in the comparator)
		//
		// TODO: use a gpu depth buffer with parallax as z-coordinate?
		// Might be overkill.
		sprites.sort_by(|a, b| b.parallax.total_cmp(&a.parallax));

		return sprites;
	}
}

pub async fn run() -> Result<()> {
	let event_loop = EventLoop::new();
	let window = WindowBuilder::new().build(&event_loop).unwrap();

	let mut gpu = GPUState::new(window).await?;
	let mut game = Game::new();

	gpu.update_starfield_buffer(&game);

	event_loop.run(move |event, _, control_flow| {
		match event {
			Event::RedrawRequested(window_id) if window_id == gpu.window().id() => {
				gpu.update();
				game.update();
				match gpu.render(&game) {
					Ok(_) => {}
					// Reconfigure the surface if lost
					Err(wgpu::SurfaceError::Lost) => gpu.resize(&game, gpu.window_size),
					// The system is out of memory, we should probably quit
					Err(wgpu::SurfaceError::OutOfMemory) => *control_flow = ControlFlow::Exit,
					// All other errors (Outdated, Timeout) should be resolved by the next frame
					Err(e) => eprintln!("{:?}", e),
				}
			}

			Event::MainEventsCleared => {
				// RedrawRequested will only trigger once unless we manually
				// request it.
				gpu.window().request_redraw();
			}

			Event::WindowEvent {
				ref event,
				window_id,
			} if window_id == gpu.window.id() => {
				match event {
					WindowEvent::CloseRequested => *control_flow = ControlFlow::Exit,
					WindowEvent::KeyboardInput {
						input:
							KeyboardInput {
								state,
								virtual_keycode: Some(key),
								..
							},
						..
					} => game.process_key(state, key),
					WindowEvent::MouseInput { state, button, .. } => {
						game.process_click(state, button);
					}
					WindowEvent::MouseWheel { delta, phase, .. } => {
						game.process_scroll(delta, phase);
					}
					WindowEvent::Resized(physical_size) => {
						gpu.resize(&game, *physical_size);
					}
					WindowEvent::ScaleFactorChanged { new_inner_size, .. } => {
						// new_inner_size is &&mut so we have to dereference it twice
						gpu.resize(&game, **new_inner_size);
					}
					_ => {}
				}
			}
			_ => {}
		}
	});
}

fn main() -> Result<()> {
	let mut raw_content = Vec::new();
	for e in WalkDir::new("content").into_iter().filter_map(|e| e.ok()) {
		if e.metadata().unwrap().is_file() {
			let c = crate::content::ContentType::from_path(e.path())?;
			match c {
				Some(c) => raw_content.push(c),
				None => continue,
			}
		}
	}

	let a = crate::content::Content::new(raw_content)?;
	println!("{:?}", a);

	//pollster::block_on(run())?;
	return Ok(());
}
