use anyhow::Result;
use std::{collections::HashMap, fs::File, io::Read, path::Path, path::PathBuf};

use super::{rawtexture::RawTexture, Texture, TextureArrayConfig};

mod fields {

	use super::{File, HashMap, Path, PathBuf, RawTexture, Read, Result, TextureArrayConfig};
	use serde::Deserialize;

	// Config is used outside of this file,
	// so it is defined in mod.rs.

	#[derive(Deserialize)]
	pub struct Index {
		pub config: TextureArrayConfig,
		pub include: Option<HashMap<PathBuf, String>>,
		pub texture: Vec<Texture>,
	}

	impl Index {
		pub fn to_sub(self) -> (TextureArrayConfig, SubIndex) {
			(
				self.config,
				SubIndex {
					include: self.include,
					texture: self.texture,
				},
			)
		}
	}

	#[derive(Deserialize)]
	pub struct SubIndex {
		pub include: Option<HashMap<PathBuf, String>>,
		pub texture: Vec<Texture>,
	}

	#[derive(Deserialize)]
	pub struct Texture {
		pub name: String,
		pub path: String,
	}

	impl Texture {
		pub fn read(
			&self,
			device: &wgpu::Device,
			queue: &wgpu::Queue,
			root: &Path,
		) -> Result<RawTexture> {
			let mut f = File::open(root.join(&self.path))?;
			let mut bytes = Vec::new();
			f.read_to_end(&mut bytes)?;
			RawTexture::from_bytes(&device, &queue, &bytes, &self.name)
		}
	}
}

pub struct TextureLoader {
	pub textures: HashMap<String, Texture>,
	pub texture_data: Vec<RawTexture>,
	pub config: TextureArrayConfig,
}

impl TextureLoader {
	const INDEX_NAME: &'static str = "_index.toml";

	pub fn load(device: &wgpu::Device, queue: &wgpu::Queue, index_path: &Path) -> Result<Self> {
		let index: fields::Index = {
			let mut texture_index_string = String::new();
			let _ = File::open(index_path.join(Self::INDEX_NAME))?
				.read_to_string(&mut texture_index_string);
			toml::from_str(&texture_index_string)?
		};

		let (config, index) = index.to_sub();
		let texture_raw =
			Self::inner_load_textures(device, queue, index, index_path, "".to_string())?;

		// TODO: validate configuration

		let mut textures = HashMap::new();
		let mut texture_data = Vec::new();
		let mut i = 0;
		for (name, data) in texture_raw {
			textures.insert(
				name,
				Texture {
					index: i,
					aspect: data.dimensions.0 as f32 / data.dimensions.1 as f32,
				},
			);
			texture_data.push(data);
			i += 1;
		}

		return Ok(Self {
			textures,
			texture_data,
			config,
		});
	}

	fn inner_load_textures(
		device: &wgpu::Device,
		queue: &wgpu::Queue,
		index: fields::SubIndex,
		texture_root: &Path,
		prefix: String,
	) -> Result<HashMap<String, RawTexture>> {
		// Load all textures
		let mut texture_raw = HashMap::new();
		for t in index.texture {
			let data = match t.read(device, queue, &texture_root) {
				Ok(r) => r,
				Err(_) => {
					println!(
						"[WARNING] Failed to load texture `{}` from `{}`: file doesn't exist.",
						if prefix.is_empty() {
							t.name
						} else {
							format!("{prefix}::{}", t.name)
						},
						texture_root.display(),
					);
					continue;
				}
			};

			if texture_raw.contains_key(&t.name) {
				println!(
					"[WARNING] Subindex in `{}` has a duplicate texture `{}`, skipping.",
					texture_root.display(),
					if prefix.is_empty() {
						t.name
					} else {
						format!("{prefix}::{}", t.name)
					},
				);
				continue;
			}

			texture_raw.insert(
				if prefix.is_empty() {
					t.name
				} else {
					format!("{prefix}::{}", t.name)
				},
				data,
			);
		}

		// Load included files
		if let Some(include) = index.include {
			for (path, sub_prefix) in include {
				let sub_root = texture_root.join(&path);
				let index: fields::SubIndex = {
					let mut texture_index_string = String::new();
					let _ = File::open(sub_root.join(Self::INDEX_NAME))?
						.read_to_string(&mut texture_index_string);
					toml::from_str(&texture_index_string)?
				};

				let sub_textures = Self::inner_load_textures(
					device,
					queue,
					index,
					&sub_root,
					if prefix.is_empty() {
						sub_prefix
					} else {
						format!("{prefix}::{sub_prefix}")
					},
				)?;
				for (s, data) in sub_textures {
					if texture_raw.contains_key(&s) {
						println!(
							"[WARNING] Subindex in `{}` has a duplicate texture `{}`, skipping.",
							sub_root.display(),
							&s,
						);
						continue;
					}
					texture_raw.insert(s, data);
				}
			}
		}

		return Ok(texture_raw);
	}
}
